
/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

% File: utilities.pl
% Author: Temur Kutsia
% Date: 30.10.2005
% This module implements various utilities

Copyright (c) 2005 Temur Kutsia

*************************************************************************/

:-module(utilities,[
          create_unique_temporary_ctx_or_seqvar/2,
          create_unique_temporary_fun_or_indvar/2,
          new_flags_for_concatenation/4,
          evaluate_flag_for_seq_var/2,
          evaluate_flag_for_ctx_var/2,
          select_branch/2,
          remove_bindings_for_temporary_variables/2,
          anonymous_to_singleton_variables_in_subst/2,
          select_equations/2,
          op(120,yfy,~>)]).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Generating unique temporary context, sequence, function and individual
variables.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxOrSeqVar),
                                      ctx_or_seq_var(TempCtxOrSeqVar)):-
        create_symbol_list_for_a_temporary_variable(CtxOrSeqVar,UniqueList),
        append(UniqueList,[95,95],List),
        name(TempCtxOrSeqVar,List).

create_unique_temporary_fun_or_indvar(fun_or_ind_var(IndOrFunVar),
                                      fun_or_ind_var(TempIndOrFunVar)):-
        create_symbol_list_for_a_temporary_variable(IndOrFunVar,UniqueList),
        append(UniqueList,[95],List),
        name(TempIndOrFunVar,List).

create_symbol_list_for_a_temporary_variable(SeqVar,UniqueList):-
        name(SeqVar,L),
        remove_last_underscores(L,L1),
        name(temp, TempL),
        append(L1,TempL,TempNameBaseList),
        name(TempNameBase,TempNameBaseList),
        gensym(TempNameBase, Unique),
        name(Unique,UniqueList).

remove_last_underscores(L,L1):-append(L1,[95,95],L),!.
remove_last_underscores(L,L1):-append(L1,[95],L).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Flag manipulation in regular constraints: Generating new flags for the
constraints that arise after splitting concatenation, and evaluating
flag values.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% New flag generation
new_flags_for_concatenation(0,_,0,0).
new_flags_for_concatenation(1,Var,0,circplus(Var,1)).

% Flag evaluation
evaluate_flag_for_seq_var(Flag,Flag):-integer(Flag),!.
evaluate_flag_for_seq_var(circplus(X,Y),Value):-
        nonempty_seq(X,V1),
        nonempty_seq(Y,V2),
        my_xor(circplus(V1,V2),Value).

evaluate_flag_for_ctx_var(Flag,Flag):-integer(Flag),!.
evaluate_flag_for_ctx_var(circplus(X,Y),Value):-
        nonempty_ctx(X,V1),
        nonempty_ctx(Y,V2),
        my_xor(circplus(V1,V2),Value).

my_xor(circplus(X,Y),V):-
        integer(X),
        integer(Y),
        !,
        V is X xor Y.
my_xor(Unevaluated,Unevaluated).

nonempty_seq(seq([]),0):-!.
nonempty_seq(S,1):-nonempty_seq_aux(S,1),!.
nonempty_seq(S,S).

nonempty_seq_aux(seq([X|_]),1):-X\=ctx_or_seq_var(_),!.
nonempty_seq_aux(seq([_|Y]),V):-nonempty_seq_aux(seq(Y),V).

nonempty_ctx(hole,0).
nonempty_ctx(C,1):-nonempty_ctx_aux(C,1),!.
nonempty_ctx(C,C).

nonempty_ctx_aux([[X|L]|_],1):-nonempty_ctx_aux([X|L],1),!.
nonempty_ctx_aux([[_|_]|Tail],V):-!,nonempty_ctx_aux(Tail,V).
nonempty_ctx_aux([H|_],1):-H\=hole,H\=ctx_or_seq_var(_),!.
nonempty_ctx_aux([_|Tail],V):-nonempty_ctx_aux(Tail,V).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Selecting a branch in the 'or' regular expression
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

select_branch(or(R1,_R2),R1).
select_branch(or(_R1,R2),R2).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
anonymous_to_singleton_variables_in_subst replaces in a substitution each
anonymous variable with a new temporary singleton variable
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

anonymous_to_singleton_variables_in_subst([],[]).
anonymous_to_singleton_variables_in_subst([X ~> RHS|Tail],[X~>RHS1|Tail1]):-
        anonymous_to_singleton_variables_in_term(RHS,RHS1),
        anonymous_to_singleton_variables_in_subst(Tail,Tail1).

anonymous_to_singleton_variables_in_term('__',Var):-!,
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(an__),Var).
anonymous_to_singleton_variables_in_term('_',Var):-!,
        create_unique_temporary_fun_or_indvar(fun_or_ind_var(an_),Var).
anonymous_to_singleton_variables_in_term(seq(L),seq(L1)):-!,
        anonymous_to_singleton_variables_in_term(L,L1).
anonymous_to_singleton_variables_in_term([X],[Term]):-!,
        anonymous_to_singleton_variables_in_term(X,Term).
anonymous_to_singleton_variables_in_term([X|Tail],[X1|Tail1]):-!,
        anonymous_to_singleton_variables_in_term(X,X1),
        anonymous_to_singleton_variables_in_term(Tail,Tail1).
anonymous_to_singleton_variables_in_term(X,X).


              

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
select_equations takes a matching problem and returns a list of all
matching equations that occur in the input, removing from it all regular
constraints.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

select_equations([],[]).
select_equations([X << Y|Tail], [X << Y |Tail1]):-!,
        select_equations(Tail,Tail1).
select_equations([_|Tail], L):-select_equations(Tail,L).







